"""service ordering

Revision ID: 6adeb5d3032e
Revises: 25f32879e0fb
Create Date: 2019-07-09 12:02:39.164203

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6adeb5d3032e'
down_revision = 'd28e52d86633'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('service_queue', sa.Column('created_on', sa.DateTime(), nullable=True))
    op.add_column('service_queue', sa.Column('modified_on', sa.DateTime(), nullable=True))
    op.add_column('service_queue', sa.Column('priority', sa.Integer(), nullable=True))
    op.create_index('ix_service_queue_modified_on', 'service_queue', ['modified_on'], unique=False)
    op.create_index('ix_service_queue_priority', 'service_queue', ['priority'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('ix_service_queue_priority', table_name='service_queue')
    op.drop_index('ix_service_queue_modified_on', table_name='service_queue')
    op.drop_column('service_queue', 'priority')
    op.drop_column('service_queue', 'modified_on')
    op.drop_column('service_queue', 'created_on')
    # ### end Alembic commands ###
